package net.deliciousreya.minecraftportal.model

import com.google.common.collect.ImmutableMap
import net.deliciousreya.minecraftportal.extensions.MINERAL_MAPPING
import net.deliciousreya.minecraftportal.proto.PortalSaveDataProtos
import org.bukkit.Material
import org.bukkit.Server
import java.io.File
import java.lang.IllegalArgumentException

class PortalDataStore() {
    val portalTypes: ImmutableMap<Material, PortalType>
    init {
        val portalTypesBuilder = ImmutableMap.builder<Material, PortalType>()
        for (mineralType in MINERAL_MAPPING.keys) {
            portalTypesBuilder.put(mineralType, PortalType(mineralType))
        }
        portalTypes = portalTypesBuilder.build()
    }

    fun onAfterChanged() {
        // Does nothing for now.
        // TODO: Save the data.
    }

    /** Marks the new portal as active, and removes the one that needs to be deactivated, if any. */
    fun activateAndReplacePortal(frame: PortalFrame): PortalFrame? {
        val type = portalTypes[frame.mineral] ?: throw IllegalArgumentException("Can't create a portal of type ${frame.mineral}")
        val result = type.addOrReplacePortal(Portal(frame, type))?.frame
        onAfterChanged()
        return result
    }

    /** Deactivates the given portal. */
    fun deactivatePortal(frame: PortalFrame): Boolean {
        val type = portalTypes[frame.mineral] ?: throw IllegalArgumentException("There are no portals of type ${frame.mineral}")
        val result = type.removePortalWithFrame(frame)
        onAfterChanged()
        return result
    }

    /** Forgets all the data without actually affecting the world. */
    fun clear() {
        for (portalType in this.portalTypes.values) {
            portalType.clear()
        }
    }

    fun loadFromProto(server: Server, proto: PortalSaveDataProtos.PortalSaveData) {
        clear()
        for (portalProto in proto.unpairedPortalsList) {
            val portal = portalProto.toPortal(server, portalTypes)
            portal.type.loadUnpairedPortal(portal)
        }
        for (portalPair in proto.pairedPortalsList) {
            val olderPortal = portalPair.older.toPortal(server, portalTypes)
            val newerPortal = portalPair.newer.toPortal(server, portalTypes)
            if (olderPortal.type != newerPortal.type) {
                throw DeserializationException("Pair had non-matching types: ${olderPortal.type} and ${newerPortal.type}")
            }
            olderPortal.type.loadPairedPortals(olderPortal, newerPortal)
        }
    }

    fun toProto(): PortalSaveDataProtos.PortalSaveData {
        val builder = PortalSaveDataProtos.PortalSaveData.newBuilder()
        for (portalType in this.portalTypes.values) {
            when {
                portalType.isEmpty -> {
                    // Nothing here to save. Just skip it.
                }
                portalType.isPaired -> {
                    builder.addPairedPortals(portalType.toProto())
                }
                else -> {
                    builder.addUnpairedPortals(portalType.getOnlyPortal.toProto())
                }
            }
        }
        return builder.build()
    }
}